import base64
import datetime
from dataclasses import dataclass
from typing import Collection, Dict, Iterable, List, Optional, Union

from cryptography.x509 import Certificate
from lxml.etree import QName, _Element as Element
from minisignxml.config import VerifyConfig
from minisignxml.errors import ElementNotFound
from minisignxml.verify import extract_verified_element_and_certificate

from .errors import (
    AudienceMismatch,
    MalformedSAMLResponse,
    ResponseExpired,
    ResponseTooEarly,
)
from .internal.constants import *
from .internal.namespaces import NAMESPACE_MAP
from .internal.saml import saml_to_datetime
from .internal.utils import find_or_raise


@dataclass(frozen=True)
class Attribute:
    name: str
    values: List[str]
    format: Optional[str]
    extra_attributes: Dict[str, str]

    @property
    def value(self) -> Optional[str]:
        return self.values[0] if self.values else None


@dataclass(frozen=True)
class Response:
    issuer: str
    name_id: str
    audience: str
    attributes: List[Attribute]
    session_not_on_or_after: Optional[datetime.datetime]
    in_response_to: Optional[str]
    certificate: Certificate

    @property
    def attrs(self) -> Dict[str, Optional[str]]:
        return {attr.name: attr.value for attr in self.attributes}


def validate_response(
    *,
    data: Union[bytes, str],
    certificate: Union[Certificate, Collection[Certificate]],
    expected_audience: str,
    signature_verification_config: VerifyConfig = VerifyConfig.default()
) -> Response:
    xml = base64.b64decode(data)
    certificates: Collection[Certificate]
    if isinstance(certificate, Certificate):
        certificates = {certificate}
    else:
        certificates = certificate
    element, certificate_used = extract_verified_element_and_certificate(
        xml=xml, certificates=certificates, config=signature_verification_config
    )
    if element.tag == QName(NAMES_SAML2_PROTOCOL, "Response"):
        assertion = find_or_raise(element, "./saml:Assertion")
    elif element.tag == QName(NAMES_SAML2_ASSERTION, "Assertion"):
        assertion = element
    else:
        raise MalformedSAMLResponse(
            "Signed element is neither a Response with an Assertion, nor an Assertion"
        )
    issuer = find_or_raise(assertion, "./saml:Issuer").text
    subject = find_or_raise(assertion, "./saml:Subject")
    name_id = find_or_raise(subject, "./saml:NameID").text
    subject_confirmation_method = find_or_raise(subject, "./saml:SubjectConfirmation")
    subject_confirmation_data = find_or_raise(
        subject_confirmation_method, "./saml:SubjectConfirmationData"
    )
    in_response_to = subject_confirmation_data.attrib.get("InResponseTo", None)
    conditions = find_or_raise(assertion, "./saml:Conditions")
    not_before = saml_to_datetime(conditions.attrib["NotBefore"])
    not_on_or_after = saml_to_datetime(conditions.attrib["NotOnOrAfter"])
    now = datetime.datetime.utcnow()
    if now < not_before:
        raise ResponseTooEarly()
    if now >= not_on_or_after:
        raise ResponseExpired()

    audience = find_or_raise(
        conditions, "./saml:AudienceRestriction/saml:Audience"
    ).text

    if audience != expected_audience:
        raise AudienceMismatch(
            received_audience=audience, expected_audience=expected_audience
        )

    raw_session_not_on_or_after = find_or_raise(
        assertion, "./saml:AuthnStatement"
    ).attrib.get("SessionNotOnOrAfter", None)

    session_not_on_or_after = raw_session_not_on_or_after and saml_to_datetime(
        raw_session_not_on_or_after
    )

    try:
        attribute_statement = find_or_raise(assertion, "./saml:AttributeStatement")
    except ElementNotFound:
        attribute_statement = None

    attributes = (
        list(gather_attributes(attribute_statement))
        if attribute_statement is not None
        else []
    )

    return Response(
        issuer=issuer,
        name_id=name_id,
        audience=audience,
        attributes=attributes,
        session_not_on_or_after=session_not_on_or_after,
        in_response_to=in_response_to,
        certificate=certificate_used,
    )


def gather_attributes(attribute_statement: Element) -> Iterable[Attribute]:
    for attribute in attribute_statement.findall("./saml:Attribute", NAMESPACE_MAP):
        values = [
            value.text
            for value in attribute.findall("./saml:AttributeValue", NAMESPACE_MAP)
        ]
        extra_attributes = {k: v for k, v in attribute.attrib.items()}
        name = extra_attributes.pop("Name")
        format = extra_attributes.pop("NameFormat", None)
        yield Attribute(
            name=name,
            values=values,
            format=format,
            extra_attributes=extra_attributes,
        )



class MiniSAMLError(Exception):
    pass


class MalformedSAMLResponse(MiniSAMLError):
    pass


class ResponseExpired(MiniSAMLError):
    pass


class ResponseTooEarly(MiniSAMLError):
    pass


class AudienceMismatch(MiniSAMLError):
    def __init__(self, *, received_audience: str, expected_audience: str):
        self.received_audience = received_audience
        self.expected_audience = expected_audience
        super().__init__()



from minisignxml.internal.namespaces import make_namespace

from .constants import *

samlp = make_namespace("samlp", NAMES_SAML2_PROTOCOL)
saml = make_namespace("saml", NAMES_SAML2_ASSERTION)

NAMESPACE_MAP = {"samlp": NAMES_SAML2_PROTOCOL, "saml": NAMES_SAML2_ASSERTION}



import datetime

from minisignxml.internal.utils import serialize_xml

from .constants import *
from .namespaces import saml, samlp


def datetime_to_saml(t: datetime.datetime) -> str:
    return t.strftime(DATE_TIME_FORMAT)


def saml_to_datetime(s: str) -> datetime.datetime:
    if "." in s:
        return datetime.datetime.strptime(s, DATE_TIME_FORMAT_FRACTIONAL)
    return datetime.datetime.strptime(s, DATE_TIME_FORMAT)


def build_saml_request(
    issuer: str, acs_url: str, request_id: str, force_reauthentication: bool
) -> bytes:
    request = samlp.AuthnRequest(
        saml.Issuer(issuer),
        samlp.NameIDPolicy(Format=NAMEID_FORMAT_UNSPECIFIED),
        ID=request_id,
        Version="2.0",
        IssueInstant=datetime_to_saml(datetime.datetime.utcnow()),
        ProtocolBinding=BINDINGS_HTTP_POST,
        AssertionConsumerServiceURL=acs_url,
    )
    if force_reauthentication:
        request.set("ForceAuthn", "true")
    return serialize_xml(request)



from lxml.etree import _Element as Element
from minisignxml.internal import utils

from .namespaces import NAMESPACE_MAP


def find_or_raise(element: Element, path: str) -> Element:
    return utils.find_or_raise(element, path, NAMESPACE_MAP)



import base64
import secrets
import zlib
from typing import Optional

from yarl import URL

from .internal.saml import build_saml_request


def get_request_redirect_url(
    *,
    saml_endpoint: str,
    expected_audience: str,
    acs_url: str,
    force_reauthentication: bool = False,
    request_id: Optional[str] = None,
    relay_state: Optional[str] = None,
) -> str:
    request_xml = build_saml_request(
        issuer=expected_audience,
        acs_url=acs_url,
        request_id=request_id or secrets.token_urlsafe(),
        force_reauthentication=force_reauthentication,
    )

    query = {
        "SAMLRequest": base64.b64encode(zlib.compress(request_xml)[2:-4]).decode(
            "utf-8"
        )
    }
    if relay_state is not None:
        query["RelayState"] = relay_state
    return str(URL(saml_endpoint).with_query(query))



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['minisaml', 'minisaml.internal']

package_data = \
{'': ['*']}

install_requires = \
['lxml>=4.4.1', 'minisignxml>=20.0', 'yarl>=1.4.2']

extras_require = \
{'docs': ['sphinx>=3.2.0,<4.0.0', 'sphinxcontrib-mermaid>=0.4.0,<0.5.0']}

setup_kwargs = {
    'name': 'minisaml',
    'version': '21.10',
    'description': 'Minimal SAML2 client',
    'long_description': "# MiniSAML\n\n\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n[![CircleCI](https://circleci.com/gh/HENNGE/minisaml.svg?style=svg)](https://circleci.com/gh/HENNGE/minisaml)\n[![Documentation Status](https://readthedocs.org/projects/minisaml/badge/?version=latest)](https://minisaml.readthedocs.io/en/latest/?badge=latest)\n\n\n\nAbsolutely minimalistic SAML 2 client. Does not support the full SAML 2 specification, on purpose.\nIt only supports requests via HTTP Redirect and responses via HTTP POST.\n\n\n## Usage\n\n\n### Create a SAML Request\n\n```python\nfrom minisaml.request import get_request_redirect_url\n\nurl = get_request_redirect_url(\n    saml_endpoint='https://your-idp.invalid/sso-endpoint/', \n    expected_audience='Your SAML Issuer', \n    acs_url='https://you.web-site.invalid/saml/acs/'\n)\n\n# This line depends on your web framework/server\nredirect_user_to_url(url)\n```\n\n### Validate and parse the SAML Response\n\n```python\nfrom minisaml.response import validate_response\n\n# This line depends on your web framework/server\nsaml_response = get_SAMLResponse_form_data_as_bytes() \n\n# Load the x509 certificate as a cryptography.x509.Certificate somehow\ncertificate = ...\n\ntry:\n    response = validate_response(data=saml_response, certificate=certificate, expected_audience='Your SAML Issuer')\nexcept:\n    handle_invalid_response_somehow()\n\n# response is a minisaml.response.Response object\n```\n",
    'author': 'Jonas Obrist',
    'author_email': 'jonas.obrist@hennge.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/HENNGE/minisaml',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



NAMES_SAML2_PROTOCOL = "urn:oasis:names:tc:SAML:2.0:protocol"
NAMES_SAML2_ASSERTION = "urn:oasis:names:tc:SAML:2.0:assertion"
NAMEID_FORMAT_UNSPECIFIED = "urn:oasis:names:tc:SAML:1.1:nameid-format:unspecified"
BINDINGS_HTTP_POST = "urn:oasis:names:tc:SAML:2.0:bindings:HTTP-POST"
DATE_TIME_FORMAT = "%Y-%m-%dT%H:%M:%SZ"
DATE_TIME_FORMAT_FRACTIONAL = "%Y-%m-%dT%H:%M:%S.%fZ"
